
; ------------------------------------------------------------
; A Z interface to ACL2/Lisp

; From zmacros4.lisp by bevier and cohen.
; Mods for parsing and execution begun 12/28/96. M. K. Smith

; For more tests from the Synergy file system
; see /cli/project/os/notes/note-100/Interface/sfs20.tex
; see /cli/project/os/notes/note-100/Model/*

; ----------------------------------------------------------------------
; Main entry points.
;
; (loadz "file") : file Z file (currently in ZTC format).  Creates
;    a database that can be used to create either executable forms
;    or theorems.
; (reset) : clears the database.

; (generate-executable) : creates the executable versions.
; (test &optional calls) : Runs test cases.
;    If CALLS is provided, simulate the series of CALLS.
;    The simulation step consist of a call followed by (STEP), which
;    copies the primed vars to their unprimed versions and clears
;    the primed vars.
;    Null CALLS case not yet implemented. The intention is that
;    if CALLS is left empty TEST will randomly select from the 
;    executable functions (choosing arguments as it needs to).

; ----------------------------------------------------------------------
; Forms expected to be produced by the parser.
;
; (base-type-decl . names)  Adds to *types*
; (free-type name branches)         *types, *constants*, *constructors* 
; (schema name decls preds)         *schemas* 
; (gendef name decls preds)         *generics* 
; (genequiv lhs schema-exp)         *generics* 
; (schemadef name-args exp)         *schemas* or *generics* 
; (axiom decls preds)       Adds preds to *axioms*, decls to *globals*, *functions*

; ----------------------------------------------------------------------
; Database components (see structures.lisp)
;
; Defines structures in various lists:
;  *types*     - basetype additions due to basic-type-decls.
;  *constants* - basetype additions due to free-types
;  *globals*   - global vars
;  *axioms*    - 
;  *schemas*   - 
;  *generics*   - 

; (generate-executable)
; Initializes executable state by defining
;    constants and globals
;    predicates for schema invariants and transitions
;    executable versions of various pfuns
;    executable images of schemas
;
; To 'execute a schema', call schema-fn with the appropriate args.
;
; Still schizoid.  We want to accomplish two distinct things.
;
; 1. Executable Z.
; 2. ACL2 checked Z. Note  #+:SMALL-ACL2-IMAGE
;
; We should go to some parsed form from which we can generate
; either.  So, the tool will let you parse in a file.  It will
; create a schema database, into which it will store the results
; of some limited type checking and catagorization of variables.  
; From this you can request further analysis or the generation 
; of executable code.

; Subset of Z:
;  Include
;   Global, Axiom, Schema, Generic

;  Types: We generate types similar to Z, but constructive.
;   Type = integer, basic_type, (powerset type), (cross . type*),
;          (schema . (name . type)*)
;   A base type of a type is the fully expanded form, containing
;   in the leaves nothing other than basic_types and integer.

; There are any number of problems translating Z specs, primary
; among them being that decorations on variables are only suggestive.
; If we are trying to specify an operation its inputs may 
; really include any schema variables.  

; One of the more useful tools we might create is one that detects
; inconsistencies in predicates.  Fuzz will detect type errors.

; Our basic model of a state is an alist that binds variables
; to values. A simple schema macro generates a defun that is a 
; predicate on such an alist.
;
; A schema macro that includes input OR output variables 
; can may generate a transition function that executes 
; the transition.  The function, G, is a multiple-valued
; function from the set of ? variables and, implicitly, state,
; to (mv . !variables)).  
; Translation depends on the form of x! and x' var predicates.
;
; 1a. (equal x! f(..)) - we just compute it.
; 1b. (equal x' f(..)) - we just compute it.
; 2a. (REL x! f(..)    - we CHOOSE.  E.g. x! = (choose x! x.TYPE, (REL x! f(..)))
; 2b. (REL x' f(..)    - we CHOOSE.  E.g. x' = (choose x' x.TYPE, (REL x' f(..)))
;    Depends on a generator for x.TYPE.
;    We do not backtrack outside of REL. So
;    (> x! 0)                    generates x! = 1
;    (and (< y! x!) (> y! 2))
;    will fail if our choose function starts at 0 and increments until (> x! 0)
;    is satisfied.
; 3. Otherwise we fail to generate a transition function.
;
; In cases 1 and 2, we return a multiple value, (mv . (!variables . state)),
; with the x' vars bound into state. 

; This file currently is a bunch of hacks that would need to get
; thoroughly reviewed.
;
; Look for the string REMARK to see some comments about
; potential problems with the code.
;
; From a Schema, S, we create:
; 
; Type checker: Are all vars of the proper type.
;  s-type (vars) -> boolean    
; Predicate checker: Are all preds satisfied. It is either an invariant, 
; for simple schemas or a delta for those describing a state change.
;  a. s-invar (state) -> boolean 
;  b. s-delta (state state' . input @ output ) -> boolean  
; Executable function: if possible, we create:
;  s-fn (input) -> (output)

; Adds S to *schemas*, and may add to *types* *constants* *globals*


; REMARKs
;  In Z, there is an implicit guard in combining schemas: all
;        variables in common must have the same types. We are currently
;        providing no help in making such a guarantee. Perhaps we should
;        generate a guard theorem.
;
; ------------------------------------------------------------

(require "zpackage" "zpackage.lisp")
(in-package "Z")

; ------------------------------------------------------------
; Preliminaries

; Define pack, pack*, match, bind, ...
(require "utilities" "/usr/cli/projects/z/ez/execute/utilities")
(require "structures" "/usr/cli/projects/z/ez/execute/structures")
(require "operators" "/usr/cli/projects/z/ez/execute/operators")
(require "zlisp-events" "/usr/cli/projects/z/ez/execute/zlisp-events")

;; State = <
;;  *types*        : ((id . type)*)     - type is structure (and fn name)
;;  *globals*      : ((id . typedesc)*) - typedesc is type descriptor
;;                   Includes functions.
;;  *constants*    : ((id . typeid)*)   - newly introduced constants (from free types)
;;  *constructors* : ((id . constructor)*) - constructor is struct (from free types)
;;                   Id will also be the function name.
;;  *functions*    : ((id . fn)*)       - fn is structure
;;                   Id will be fn name, potentially defined in a number of ways.
;;                   It may use id-alist to dyamically change def.
;;                   It may be a constructed function created from an axiom.
;;                   It may be the transition function assoc with a schema
;;                    whose vars are decorated with ? and/or !.
;;  *axioms*       : (expr*) - Union of predicates from all axiom boxes.
;;  *schemas*      : ((id . schema)*) - schema is structure
;; >

(defvar *types*        nil "An alist of (id . @type)")
(defvar *globals*      nil "An alist of (id . typedesc)")
(defvar *constants*    nil "An alist of (id . typeid)")
(defvar *constructors* nil "An alist of (id . @constructor)")
(defvar *functions*    nil "An alist of (id . @function)")
(defvar *schemas*      nil "An alist of (id . @schema)")
(defvar *generics*     nil "An alist of (id . @schema)")
(defvar *axioms*       nil "A list of predicates")

(defun zglobal-p (name)
  (or (zglobalp name) (assoc name *globals*)))

(defun zglobal-getv (v)
  (cond ((zglobalp v) (zglobal-value v))
	((assoc v *globals*)
	 (zglobal-getv (cdr (assoc v *globals*))))
	(t (error (format nil "~%~s is not a global variable (in getv)." v)))))

(defun zglobal-setv (v value)
  (cond ((zglobalp v) (setf (zglobal-value v) value))
	((assoc v *globals*)
	 (zglobal-setv (cdr (assoc v *globals*)) value))
	(t (error (format nil "~%~s is not a global variable (in setv)" v)))))

(defun zconstant-p(name) (assoc name *constants*))
(defun zconstructor-p(name) (assoc name *constructors*))
(defun zfunction-p(name) (or (symbol-function name)
			     (assoc name *functions*)))
(defun schema-p  (name)
  (if (schemap name)
      name
      (let ((g (assoc name *schemas*)))
	(and g (eq (schema-kind (cdr g)) 'schema) (cdr g)))))

(defun generic-p (name)
  (let ((g (assoc name *schemas*)))
    (and g (member (schema-kind g) '(generic-schema generic-constant)))))

(defvar *current-schema* nil)

(defun ref (name)
  (let ((s (schema-p name)))
    (or s (getv name *current-schema*))))

(defun basic-decl (type &rest names) 
  ;; This is really bogus.  The parser returns both variables and schemas 
  ;; wrapped up in basic-decl.
  (cond ((null type)
	 (if (and names (null (cdr names)))
	     (ref (car names))
	     (error (format nil "(basic-decl ~s ~s) invoked incorrectly" type names))))
	((and names (null (cdr names)))
	 (ref (car names)))
	(t (error (format nil "(basic-decl ~s ~s) invoked incorrectly" type names)))))

;; ------------------------------------------------------------------------
;; Input forms : Must remain in synch with parser.

;; Doing (loadz "file.z") or (load "file.lisp")
;; adds to the various global vars as indicated below.

;; Adds to *types*
(defmacro base-type-decl (&rest names) `(def-base-types ',names))

;; Adds to *types, *constants*, *constructors* 
(defmacro free-type (name branches) `(def-free-type ',name ',branches))

;; Adds to *schemas* 
(defmacro schema (name decls preds) `(def-schema-box ',name ',decls ',preds))

;; Adds to *generics* 
(defmacro gendef (name decls preds) `(def-generic-box ',name ',decls ',preds))

;; Adds to *generics* 
(defmacro genequiv (lhs schema-exp) `(def-generic ',lhs ',schema-exp))

;; Adds to *schemas* or *generics* 
(defmacro schemadef (name-args exp) `(def-schema ',name-args ',exp))

;; Adds preds to *axioms*, decls to *globals*, *functions*
(defmacro axiom  (decls preds) `(def-axiom  ',decls ',preds))

;; ------------------------------------------------------------------------
;; Operations: 
;; 
;; (generate-executable) - Generate executable functions
;; 
;; (generate-theorems)   - Generate defuns and theorms suitable for ACL2
;;
;; (generate-test)       - Generate a test function (a step function, with one
;;                         case per operation)

(defun generate-executable ()
  (def-executable-types *types*)
  (def-executable-vars *globals*)  
  (def-executable-constants *constants*)  
  (def-executable-schemas *schemas*)   
  ; (def-function-fns *functions*)
  ; (def-invariant-fns *functions*)
  )

;; Events:

(defun generate-theorems ()
  (def-type-events *types*)
  (def-global-events *globals*)  
  (def-constant-events *constants*)  
  (def-function-events *functions*)
  (def-schema-events *schemas*)   
  (def-invariant-events *functions*))

(defun generate-test ()
  (generate-executable)
  (def-test-function *schemas*))

; ============================================================
; I. TYPES
; ------------------------------------------------------------

(defparameter *types*
  (list (cons 'Z (make-ztype :name 'Z
			    :kind 'integer
			    :basetype 'Z
			    :fun 'integerp
			    :firstchoice 0))
	(cons 'N (make-ztype :name 'N
			    :kind 'integer
			    :basetype 'Z
			    :fun 'integerp
			    :firstchoice 0))
	(cons 'N1 (make-ztype :name 'N1
			     :kind 'integer
			     :basetype 'Z
			     :fun 'integerp
			     :firstchoice 1))))

;; We keep a hashtable of type -> basetype
;; E.g. (setf (gethash '(pfun z z) *basetypes*) '(cross z z))
(defvar *basetypes* (make-hash-table :test #'equal))

(defun basetype-of-type (term)
  (cond ((assoc term *types*) term)
	((gethash term *basetypes*))
	(t (let ((base (btype term)))
	     (sethash term *basetypes* base)))))

(defun btype (type)
  ;; Basetype is an expanded form.  
  ;; We can use equal to compare base types.
  (cond ((member type '(z n1 integer natural)) 'z)
	((symbolp type)
	 (if (assoc type *types*)
	     type
	     (error (format nil "~%Warning: ~s is not a type (btype member)" type))))
	((consp type)
	 (cond ((member (car type) '(cross rel pfun tfun pinj inj psurj surj bij ffun finj))
		(cons 'cross (mapcar #'btype (cdr type))))
	       ((member (car type) '(powerset p1 f f1))
		(list 'powerset (btype (cadr type))))
	       ((eq (car type) 'domain)
		(cadr (btype (cadr type))))
	       ((eq (car type) 'range)
		(cadr (btype (caadr type))))
	       ((member (car type) '(seq seq1 iseq))
		(list 'cross 'z (btype (cadr type))))
	       ((eq (car type) 'bag)
		(list 'cross (btype (cadr type)) 'z))
	       ((eq (car type) 'signature)
		;; The cdr of type has been sorted by name.
		;; We have not checked for duplicate names.
		;; Do we need a no-duplicates here?
		;; Followed by a check for dup names, indicating type clash?
		(cons 'signature (mapcar (f-l (pair) (cons (car pair) (btype (cdr pair))))
					 (cdr type))))
	       ;; ???
	       ;; ((member (car type) '(set)) ;; (set type)
	       ;;  (btype (nth 1 type)))
	       ((eq (car type) 'parens) (btype (cadr type)))
	       (t (error "Bad type expression ~a" type))))
	(t (warning "~%~S is not a type (btype)." type))))

(defun lexorder (x y)
  ;; 1 < #\a < "a" < 'a < nil < '(a)
  (cond ((numberp x)
	 (cond ((numberp y) (< x y))
	       ;; Otherwise
	       (t t)))
	((characterp x)
	 (cond ((numberp y) nil)
	       ((characterp y) (char< x y))
	       (t t)))
	((stringp x)
	 (cond ((stringp y) (string-lessp x y))
	       ((or (symbolp y) (consp y)) t) ; (null y) redundant
	       ;; number, char
	       (t t)))
	((symbolp x)
	 (cond ((null x)
		(if (consp y) t nil))
	       ((symbolp y) (string-lessp (symbol-name x) (symbol-name y)))
	       ((or (null y) (consp y)) t)
	       ;; number, char, string
	       (t nil)))
	((consp x)
	 (cond ((not (consp y)) nil)
	       ((consp y) (lexorder (car x) (car y)))
	       (t nil)))))

(defun vars-signature (vars)
  (btype (cons 'signature (sort (copy-list vars) #'lexorder))))

(defvar empty nil)

(defun true (&rest l) (declare (ignore l)) t)

(defun def-base-types (l) (mapc (f-l (x) (add-type x 'base x nil (pack x '|1|))) l))

(defun def-executable-types    (types) (mapc (function fn-type) types))

(defun add-type (name kind base function first)
  (setq *types*
	(append *types*
		(list (cons name (make-ztype :name name
					    :kind kind
					    :basetype base
					    :fun function
					    :firstchoice first))))))

(defun fn-type (x)
  (if (consp x) (setq x (cdr x)))
  (let* ((name (ztype-name x))
	 (kind   (ztype-kind x))
	 ;; (def (ztype-def x))
	 (function (ztype-fun x))
	 (constructors (ztype-constructors x))
	 (constants (ztype-constants x)))
    (cond ((eq kind 'base)
	   (cond ((ztype-constants x)
		  (setf (symbol-function name)
			`(lambda (x) (member x ',(ztype-constants x) :test #'equal))))
		 (t (setf (symbol-function name)
			  `(lambda (x) (declare (ignore x)) t)))))
	  (function
	   (setf (symbol-function name)
		 `(lambda (x) (funcall ',function x))))
	  ((eq kind 'free)
	   (cond ((and constructors constants)
		  (setf (symbol-function name)
			`(lambda (x)
			   (or (member x ',constants)
			       ,@(zconstructor-check-fn constructors)))))
		 (constants 
		  (setf (symbol-function name)
			`(lambda (x) (member x ',constants))))
		 (t (error (format nil "Bad free type defun, ~s" x)))))	  
	  ;; (def (setq body `(ztype-check x ',def)))
	  (t (error (format nil "Bad type defun, ~s" x))))))

(defun typename-p (name) (assoc name *types*))

(defun ztype-p (type) 
  (cond ((symbolp type)  (assoc type *types*))
	((consp type)
	 (cond ((member (car type) '(cross rel pfun tfun pinj inj psurj surj bij ffun finj))
		(every #'type-p (cdr type)))
	       ((member (car type) '(powerset p1 f f1 seq seq1 iseq bag))
		(ztype-p (cadr type)))
	       ((eq (car type) 'bindingp)
		(and (no-duplicates (mapcar #'car (cdr type)))
		     (every (f-l (x) (and (symbolp (car x))
					  (ztype-p (cdr x))))
			    (cdr type))))
	       (t nil)))
	(t nil)))

; ------------------------------------------------------------
; I.1. FREE-TYPES
;      list ::= NULL | cons(list list)
; ------------------------------------------------------------

(defun def-free-type (name branches)
  (let* ((constants (ident-constants branches))
	 (constructors (ident-constructors name branches constants)))
    (sloop for x in constants
	   ;; x is the constant, name is its type.
	   do (setassoc *constants* x name "~%Warning: Redefining type of constant ~s"))
    (setassoc *types* name
	      (make-ztype :name name
			 :kind 'free
			 :firstchoice
			 (if constants
			     (car constants)
			     (format t "~%Warning: Free type, ~s, with no constants" name))
			 :constants constants
			 :constructors constructors))))

;; Free type defuns are may by defun-type-fn.
;; So are events.

(defun ident-constants (b)
  (mapcan (f-l (x)
	       (if (and (consp x) (eq (car x) 'constant) (symbolp (cadr x)))
		   (list (cadr x))
		   nil))
	  b))

(defun ident-constructors (name b constants)
  (mapcan (f-l (x) (if (and (consp x) (eq (car x) 'constructor))
		       (list (build-zconstructor-fn name (cdr x) constants))
		       nil))
	  b))


(defun ident-recognizers (name b constants)
  (mapcon (f-l (x) (if (not (symbolp x))
		       (list (build-recognizer-term name x constants))
		       nil))
	  b))

(defun build-zconstructor-fn (name x constants)
  ;; NOTE: the contructor does not check its args.
  (declare (ignore name constants))
  (let* ((label (car x))
	 (args  (get-args (cadr x)))
	 (body  `(list ',label ,@args)))
    (if (fboundp label)
	(format t "~%Warning: Redefining ~s" label))
    (setf (symbol-function label) `(lambda ,args ,body))))

(defun zconstructor-check-fn (fn)
  ;; FN of form (lamba (args) (list 'atom a1 .. an))
  (let ((term (cdr (caddr fn))))
    (if (not (consp term))
      `(eq ',term x)
      `(and (consp x)
	    (eq (car x) ,(car term))
	    (eql (length x) ,(length term))))))
	    

(defvar *arglist* '(x1 x2 x3 x4 x5 x6 x7 x8 x9))

(defvar *max-arg-count* 9)

(defun firstn (n l)
  (cond ((<= n 0) nil)
	((null l)
	 (let ((new (pack 'x (incf *max-arg-count*))))
	   (setq *arglist* (append *arglist* (list new)))
	   (cons new (firstn (- n 1) nil))))
	(t (cons (car l) (firstn (- n 1) (cdr l))))))

(defun get-args (x)
  (cond ((and (consp x) (eq (car x) 'cross)) (firstn (length (cdr x)) *arglist*))
	(t (list 'x))))

(defun get-types (x)
  (cond ((and (consp x) (eq (car x) 'cross)) (cdr x))
	(t (list x))))

(defun build-recognizer-term (name x constants)
  ;; NOTE: takes a single arg that must be a list.
  ;; X is of the form (label type)
  ;; If X in nary, then type is of the form (cross . types)
  `(and (not (symbolp x))
	(eq (car x) ',(car x))
	,(if (and (consp (cadr x)) (eq (car (cadr x)) 'cross))
	     (recognize 'x (cdr (cadr x)) name constants)
	     (recognize 'x x name constants))))

(defun recognize (arg x name constants)
  (cond ((symbolp x)
	 (cond ((member x constants) `(equal ,arg ',x))
	       ((eq x name) `(,name ,arg))
	       ((assoc x *types*) `(,x ,arg))
	       (t (error (format nil "Bad type ref, ~a, in def-ident" x)))))
	(t `(and ,@(sloop for i from 0 to (- (length x) 1)
			  collect (recognize (list 'nth i arg) (nth i x)
					     name
					     constants))))))



;; -------------------------------------------------------------------
;; Schemas and schema related function naming.

(defvar *schema-count* 0)

(defun new-schema-name (generic) (pack generic '- (incf *schema-count*)))

(defun ztype-fn (name)       (pack name '-type))
(defun invar-fn (name)      (pack name '-invar))
(defun delta-fn (name)      (pack name '-delta))
(defun executable-fn (name) (pack name '-fn))

(defun axiom-fn (name)      (pack name '-axiom))
(defun theorem-fn (name)    (pack name '-theorem))
(defun disjoint-fn (name)   (pack name '-disjoint))
(defun consistency-fn (name)(pack name '-consistent))

(defun pfun-typename (x y)  (pack x '- y '-alist))
(defun power-typename (x)   (pack x '-list))
(defun seq-typename (x)     (pack x '-list))

(defun cross-typename (l)
  (if (symbolp l)
      'cross
      (pack (car l) '- (cross-type-name (cdr l)))))

(defun no-duplicates (l &key (test #'equal))
  (remove-duplicates l :test test))

;; (sloop for tail on l
;;     when (not (member (car tail) (cdr tail) :test test))
;;     collect (car tail))

(defun conjuncts (term)
  (cond ((atom term) (list term))
	((eq (ffn-symb term) 'and)
	 (sloop for x in (ffn-args term)
	   append (conjuncts x)))
	(t (list term))))

(defun andlist (l)
  (let ((l1 (sloop for term in l append (conjuncts term))))
    (cond ((atom l1) 't)
	  ((atom (cdr l1)) (car l))
	  (t (cons 'and l1)))))

; ----------------------------------------
; Variables in a context.
; ----------------------------------------

;; The *state* is just *schemas*, where any occurence of S' has
;; resulted in the addition of S' to *schemas*.

;; !!! We do not yet handle hiding or renaming of variables.
;; Except insofar as locals hide globals.

(defun defv-local (v &optional (c *current-schema*))
  "Is V defined in this context (but not globally)?"
  (let (x)
    (cond ((schemap c) 
	   (cond ((assoc v (schema-locals c)) t)
		 (t (some (f-l (y) (defv-local v y)) (schema-schemas c)))))
	  ;; Is C a schema name
	  ((setq x (schema-p c)) (defv-local v x))
	  (t nil))))

(defun defv (v &optional (c *current-schema*))
  "Is V defined in this context or globally?"
  (or (defv-local v c) (zglobal-p v)))
	       
(defun getv (v &optional (c *current-schema*))
  "Get the value of V in context C"
  (if (not (defv v c))
      (error (format nil "Undefined variable, ~a, in context ~a"
		     v (if (schemap c) (schema-name c) c)))
      (let ((values (getvs v c)))
	(cond ((and (consp values) (null (cdr values))) (car values))
	      ((null values)
	       (if (zglobal-p v)
		   (zglobal-getv v)
		   ;; Unitialized
		   nil))
	      ((cdr values)
	       (error (format nil "Variable ~a bound to multiple values in context ~a"
			      v (if (schemap c) (schema-name c) c))))
	      ;; Should never happen
	      (t (error (format nil "(GETVS ~a ~a) returned non-list"
				v (if (schemap c) (schema-name c) c))))))))

(defun getvs (v c) "Get ALL of the non-global values of V in context C"
  (let (x values)
    (cond ((schemap c) 
	   (if (assoc v (schema-locals c))
	       (let ((pair (assoc v (schema-bindings c))))
		 (if pair (setq values (list (cdr pair))) nil)))
	   (setq values (append values
				(sloop for s in (schema-schemas c)
				       appending (getvs v s))))
	   (remove-duplicates values :test #'equal))
	  ;; Is C a schema name
	  ((setq x (schema-p c)) (getvs v x))
	  (t nil))))

(defun setv (v value &optional (c *current-schema*))
  "Set ALL of the values of V in this context"
  (if (defv v c)
      (setv2 v value c)
      (error (format nil "Variable ~a non-existent in global or local context ~a"
		     v (if (schemap c) (schema-name c) c)))))

(defun setv2 (v value c) "Set the value of V in this context"
  (let (x)
    (cond ((and c (defv-local v c))
	   (cond ((schemap c) 
		  (if (assoc v (schema-locals c))
		      (let ((bindings (schema-bindings c)))
			(setf (schema-updated c) t)
			(setf (schema-bindings c)
			      (addbinding bindings v value))))
		  (sloop for s in (schema-schemas c)
			 do (setv2 v value s))
		  value)
		 ;; Is C a schema name
		 ((setq x (schema-p c)) (setv2 v value x))
		 (t nil)))
	  ((zglobal-p v) (zglobal-setv v value))
	  (t nil))))

(defvar bindings-printed nil)

(defun print-bindings (s)
  (let (bindings-printed)
    (declare (special bindings-printed))
    (format t "~%----")
    (cond ((schemap s) (print-bindings2 s))
	  ((schema-p s) (print-bindings2 (schema-p s)))
	  (t (format t "~%Not a schema: ~s.~%" s)))
    (format t "~%----")))

(defun print-bindings2 (s)
  (let* ((name (schema-name s))
	 (bindings (schema-bindings s))
	 (schemas (schema-schemas s)))
    (cond ((member s bindings-printed) nil)
	  (bindings
	   (format t "~%~s" name)
	   (mapc (function (lambda (x) (format t "~%  ~s = ~s" (car x) (cdr x))))
		 bindings)
	   (setq bindings-printed (cons s bindings-printed))
	   (mapc (function (lambda (x) (print-bindings2 (schema-p x))))
		 schemas))
	  (t (setq bindings-printed (cons s bindings-printed))
	     (mapc (function (lambda (x) (print-bindings2 (schema-p x))))
		   schemas)))))

(defmacro union! (a b) `(union ,a ,b :test #'equal))
(defmacro set-difference!  (a b) `(set-difference ,a ,b :test #'equal))

(defmacro id-data (id l)
  (let ((l2 (pseudo-eval l))
	(type (cdr (assoc id *types*)))
	(global (cdr (assoc id *globals*))))
    (cond (type
	   (sloop for x in l2
		  do (setassoc *constants* x id "~%Warning: Redefining type of constant data ~s"))
	   ;; NOTE: The constants will be used to define the function ID in FN-TYPE.
	   (setf (ztype-constants type) (union! (ztype-constants type) l2))
	   ;; We overwrite the firstchoice
	   (setf (ztype-firstchoice type) (car l2)))
	  (global (set id l2))
	  (t (format *terminal-io*
		     "~%!!! Attempt to provide data for ~a, which 
is neither a type nor a global" id))))
    (format nil "Type ~s data" id))

(defmacro schema-data (s &rest l)
  (let ((c (schema-p s)))
    (mapc (function
	   (lambda (id-data)
	     (let ((id (cadr id-data))
		   (l2 (pseudo-eval (caddr id-data))))
	       (setv id l2 c))))
	  l)
    (format nil "Schema ~s data" s)))

(defun pseudo-eval (l)
  (cond ((atom l) l)
	((eq (car l) 'list) (mapcar #'pseudo-eval (cdr l)))
	((eq (car l) 'set*) (mapcar #'pseudo-eval (cdr l)))
	((eq (car l) 'tuple)(mapcar #'pseudo-eval (cdr l)))
	((eq (car l) 'cons) (cons (pseudo-eval (cadr l)) (pseudo-eval (caddr l))))
	((eq (car l) 'mapsto) (cons (pseudo-eval (cadr l)) (pseudo-eval (caddr l))))
	((eq (car l) 'parens) (pseudo-eval (cadr l)))
	(t (cons (pseudo-eval (car l)) (pseudo-eval (cdr l))))))

; ----------------------------------------
; Substitution
; ----------------------------------------

(defun zsubst-alist (alist term)
  (let ((dog term))
    (sloop for pair in alist 
      do (setq dog (zsubst-expr (cdr pair) (car pair) dog)))
    dog))

(defun variablep (x) (symbolp x))

(defun zsubst-expr (new old form)
  (cond ((variablep old) (zsubst-var new old form))
        ((fquotep old) (zsubst-expr-error1 old))
        (t (zsubst-expr1 new old form))))

(defun zsubst-expr-error1 (old)
  (illegal nil "Attempt to substitute for the explicit constant 
~x0. The substitution functions were optimized to disallow this."
	   (list (cons #\0 old))))

(defun zsubst-expr-lst (new old lst)
  (sloop for x in lst collect (zsubst-expr new old x)))

(defun zsubst-expr1 (new old form)
  (cond ((equal old form) new)
	((variablep form) form)
	((fquotep form) form)
	((match form (exists x y))
	 (bind form (exists x y)
	       (cons-term (ffn-symb form)
			  (list x (zsubst-expr1 new old y)))))
	(t (cons-term (ffn-symb form)
		      (sloop for arg in (ffn-args form)
			     collect (zsubst-expr1 new old arg))))))

(defun zsubst-var (new old form)
  (cond ((variablep form)
	 (cond ((eq form old) new)
	       (t form)))
	((fquotep form) form)
	((match form (exists x y))
	 (bind form (exists x y)
	       (if (member old x :test #'eq)
		   y
		   (cons-term (ffn-symb form)
			      (list x (zsubst-var new old y))))))
	(t (cons-term (ffn-symb form)
		      (sloop for arg in (ffn-args form)
			     collect (zsubst-var new old arg))))))

; ------------------------------------------------------------
; global variables
; ------------------------------------------------------------

; The following variable defines the symbol used as the formal
; argument to all schema predicates.

(defvar *schema-parameter* 'st)

; To make symbols recognizable to nqthm, we had to pick decorations
; from the set: $^&*_-+=~{}?<>
; Acl2 is not so picky, though x' will print as |x'|.

(defvar *prime-decoration* '|'|)
(defvar *input-decoration* '?)
(defvar *output-decoration* '!)

(defvar *prime-decoration-char* (char (symbol-name *prime-decoration*) 0))
(defvar *input-decoration-char* (char (symbol-name *input-decoration*) 0))
(defvar *output-decoration-char* (char (symbol-name *output-decoration*) 0))

(defvar *decoration-chars* (list *prime-decoration-char*
				 *input-decoration-char*
				 *output-decoration-char*))

(defun make-schema-name (schema-name)
  (cond ((symbolp schema-name) schema-name)
	((eq (car schema-name) 'delta) (pack (cadr schema-name) '- 'delta))
	((eq (car schema-name) 'xi)    (pack (cadr schema-name) '- 'xi))
	(t (make-schema-name (car schema-name)))))

; ------------------------------------------------------------
; DEF-SCHEMA
;
; NOTE: see DEF-SCHEMA-BOX for the standard boxed schema.  
; DEF-SCHEMA accounts for schema composition.
;
; We create several functions for schema, S.
; S-type (vars) -> boolean: checks type-correctness of vars of S.
; S-axiom (vars) -> boolean: checks that vars satisfy predicates of S.
; S (state) -> boolean: ensures vars are bound in state and of the
;   right type.  Then checks that S-axiom holds.

; Problem: We use schema-vars in 2 forms, as a list of names
; or as alist of (name.type) pairs.  Regularize this by applying var-names
; to schema-vars as needed.

; ------------------------------------------------------------

(defmacro def-schema (schema-name schexp)
  (if (and (consp schema-name) (cdr schema-name))
      (def-generic (car schema-name) (cadr schema-name) schexp)
      (let ((name (make-schema-name schema-name))
	    (vars (schexp-vars schexp))
	    ;; existential vars
	    (existentials (schexp-existentials schexp))
	    ;; pred is a single expression
	    (pred (schexp-axiom schexp)))
	`(progn
	   (setq *schemas* (cons (cons ',name
				       (make-schema :name ',name
						    :kind 'schema
						    :vars ',vars
						    :signature ',(vars-signature vars)
						    :exists ',existentials
						    :preds ',pred))
				 *schemas*))
	   (defun ,(ztype-fn name) ,(var-names vars)
	     ,(andlist (mapcar (f-l (pair) `(ztype-check ,(car pair) ',(cdr pair))) vars)))
	   ,(if (member 'exists (all-fnnames pred)) 
		`(defun-sk ,(axiom-fn name) ,(var-names vars) (exists ,existentials ,pred))
		`(defun ,(axiom-fn name) ,(var-names vars) ,pred))
	   ;; self was ,*schema-parameter*
	   (defun ,name (self)
	     (and (all-boundp (quote ,(var-names vars)) self)
		  (,(ztype-fn name)
		   ,@(sloop for var in (var-names vars)
			    collect `(getv (quote ,var) self)))
		  (,(axiom-fn name)
		   ,@(sloop for var in (var-names vars)
			    collect `(getv (quote ,var) self)))))))))

(defun defined-schema (schema-name)
  "Name must be the decorated name of a defined schema."
  (or (cdr (assoc schema-name *schemas*))
      (let ((s (undecorate-var schema-name)))
	(if (assoc s *schemas*)
	    (create-decorated-schema s (var-decoration schema-name))
	    (error (format nil "No root schema for ~s" schema-name)))))
  schema-name)

(defun create-decorated-schema (name d)
  ;; Name is the parent. D the decoration .
  (let* ((newname (decorate-var name d))
	 (s (cdr (assoc name *schemas*)))
	 new
	 (kind (schema-kind s))
	 (decls (schema-decls s))
	 (locals (schema-locals s))
	 (vars (schema-vars s))
	 (schemas (schema-schemas s))
	 (signature (schema-signature s))
	 (preds (schema-preds s))
	 (functions (schema-functions s))
	 (input  (schema-input s))
	 (output (schema-output s))
	 (exists (schema-exists s))
	 vars*)
    (declare (ignore decls kind))
    (setq new (make-schema :name newname
			   :kind 'schema
			   :locals (decorate-vars locals d)
			   :vars (setq vars* (decorate-vars vars d))
			   :schemas (mapcar #'defined-schema
					    (decorate-vars schemas d))
			   :signature (cons 'signature
					    (decorate-vars (cdr signature) d))
			   :preds (zsubst-alist (pairlis (var-names vars)
							 (var-names vars*)) preds)
			   :functions (decorate-vars functions d)
			   :input  (decorate-vars input d)
			   :output (decorate-vars output d)
			   :exists (decorate-vars exists d)
			   ))
    (setq *schemas* (cons (cons newname new) *schemas*))
    (cond ((eq *prime-decoration* d)
	   (setf (schema-next s) newname)
	   (setf (schema-prev new) name)))
    newname))

(defun sch-namep (x)
  (cond ((consp x)
	 (and (equal (length x) 2)
	   (member (car x) '(delta xi))
	   (sch-namep (cadr x))))
	((assoc x *schemas*))
	((primed-varp x)
	 (sch-namep (remove-one-decoration x)))
	(t nil)))

(defun schexp-vars (schexp)
  (cond ((sch-namep schexp)
	 (cond ((symbolp schexp)
		(if (var-decoratedp schexp)
		    (decorate-vars (sch-vars (undecorate-var schexp))
				    (var-decoration schexp))
		    (sch-vars schexp)))
	       ((or (match schexp (DELTA x))
		    (match schexp (XI x)))
		(let ((vars (schexp-vars (cadr schexp))))
		  (append vars (decorate-vars vars *prime-decoration*))))))
	((consp schexp)
	 (cond;; a kludge for def-lemma
	  ((match schexp (SCHEMA x) (schexp-vars x)))
	  ((match schexp (NOT x)    (schexp-vars x)))
	  ((match schexp (AND x y)  (union (schexp-vars x) (schexp-vars y))))
	  ((match schexp (OR x y)   (union (schexp-vars x) (schexp-vars y))))
	  ((match schexp (IMPLIES x y) (union (schexp-vars x) (schexp-vars y))))
	  ((match schexp (IFF x y) (union (schexp-vars x) (schexp-vars y))))
	  ((match schexp (HIDE x y)
		  ;; y is a list of variables
		  ;; BUG.  The first form will return name.type pairs.
		  (hide-vars (schexp-vars x) y)))
	  ((match schexp (RESTRICT x y)
		  (let ((xvars (schexp-vars x))
			(yvars (schexp-vars y)))
		    (schexp-vars `(hide (and ,x ,y)
					,(set-difference xvars yvars))))))
	  ((match schexp (EXISTS x y)
		  (schexp-vars `(hide (and ,x ,y) ,(schexp-vars x)))))
	  ((match schexp (PIPE x y)
		  (let ((qvars (piped-varnames (schexp-vars x) 
					       (schexp-vars y))))
		    (union (schexp-vars `(hide ,x ,(decorate-vars qvars 
								   *output-decoration*)))
			   (schexp-vars `(hide ,y ,(decorate-vars qvars 
								   *input-decoration*)))))))
	  ((match schexp (SEMI x y)
		  (let* ((xvars (schexp-vars x))
			 (yvars (schexp-vars y))
			 (qvars (sequenced-varnames xvars yvars)))
		    (union (schexp-vars `(hide ,x ,(decorate-vars qvars *prime-decoration*)))
			   (schexp-vars `(hide ,y ,qvars))))))
	  ((match schexp (RENAME x y)
		  ;; Note order change from input form of new / old.
		  ;; y is a list (... (oldvar-i newvar-i) ...)
		  ;; It's possible for the user to rename two vars to the same one.
		  ;; Should this be an error?
		  (rename-vars (schexp-vars x) y)))
	  ((match schexp (PRE x)
		  (let ((vars (schexp-vars x)))
		    (schexp-vars
		     `(hide ,x
			    ,(append (vars-decorated-with vars *output-decoration*)
				     (vars-decorated-with vars *prime-decoration*)))))))
	  (t (error "Bad schema expression ~a" schexp))))
	(t (error "Bad schema expression ~a" schexp))))

(defun hide-vars (vars names)
  (cond ((not (consp vars)) vars)
	((consp (car vars))
	 (if (member (caar vars) names)
	     (hide-vars (cdr vars) names)
	     (cons (car vars) (hide-vars (cdr vars) names))))
	((member (car vars) names)
	 (hide-vars (cdr vars) names))
	(t (cons (car vars) (hide-vars (cdr vars) names)))))

(defun rename-vars (vars pairs)
  ;; pairs of form (old new)
  (let (rename)
    (cond ((not (consp vars)) vars)
	  ((consp (car vars))
	   (setq rename (assoc (caar vars) pairs))
	   (if rename
	       (cons (cons (cadr rename) (cdar vars))
		     (rename-vars (cdr vars) pairs))
	       (cons (car vars) (rename-vars (cdr vars) pairs))))
	  ((setq rename (assoc (car vars) pairs))
	   (cons (cadr rename) (rename-vars (cdr vars) pairs)))
	  (t (cons (car vars) (rename-vars (cdr vars) pairs))))))

(defun schexp-existentials (schexp)
  (cond ((sch-namep schexp)
	 (cond ((symbolp schexp)
		(if (var-decoratedp schexp)
		    (decorate-vars (sch-exists (undecorate-var schexp))
				    (var-decoration schexp))
		    (sch-exists schexp)))
	       ((or (match schexp (DELTA x))
		    (match schexp (XI x)))
		(let ((vars (schexp-exists (cadr schexp))))
		  (append vars (decorate-vars vars *prime-decoration*))))))
	((consp schexp)
	 (cond;; a kludge for def-lemma
	  ((match schexp (SCHEMA x) (schexp-exists x)))
	  ((match schexp (NOT x)    (schexp-exists x)))
	  ((match schexp (AND x y)  (union (schexp-exists x) (schexp-exists y))))
	  ((match schexp (OR x y)   (union (schexp-exists x) (schexp-exists y))))
	  ((match schexp (IMPLIES x y) (union (schexp-exists x) (schexp-exists y))))
	  ((match schexp (IFF x y) (union (schexp-exists x) (schexp-exists y))))
	  ((match schexp (HIDE x y)
		  ;; y is a list of variables
		  (set-difference (schexp-exists x) y)))
	  ((match schexp (RESTRICT x y)
		  (let ((xvars (schexp-exists x))
			(yvars (schexp-exists y)))
		    (schexp-exists `(hide (and ,x ,y)
					,(set-difference xvars yvars))))))
	  ((match schexp (EXISTS x y)
		  (schexp-exists `(hide (and ,x ,y) ,(schexp-exists x)))))
	  ((match schexp (PIPE x y)
		  (let ((qvars (piped-varnames (schexp-exists x) 
					       (schexp-exists y))))
		    (union (schexp-exists `(hide ,x ,(decorate-vars qvars 
								   *output-decoration*)))
			   (schexp-exists `(hide ,y ,(decorate-vars qvars 
								   *input-decoration*)))))))
	  ((match schexp (SEMI x y)
		  (let* ((xvars (schexp-exists x))
			 (yvars (schexp-exists y))
			 (qvars (sequenced-varnames xvars yvars)))
		    (union (schexp-exists `(hide ,x ,(decorate-vars qvars *prime-decoration*)))
			   (schexp-exists `(hide ,y ,qvars))))))
	  ((match schexp (RENAME x y)
		  ;; y is a list (... (newvar-i oldvar-i) ...)
		  ;; It's possible for the user to rename two vars to the same one.
		  ;; Should this be an error?
		  (let ((vars (schexp-exists x))
			(old (sloop for pair in y collect (cadr pair)))
			(new (sloop for pair in y collect (car pair))))
		    (union (set-difference vars old)
			   new))))
	  ((match schexp (PRE x)
		  (let ((vars (schexp-exists x)))
		    (schexp-exists
		     `(hide ,x
			    ,(append (vars-decorated-with vars *output-decoration*)
				     (vars-decorated-with vars *prime-decoration*)))))))
	  (t (error "Bad schema expression ~a" schexp))))
	(t (error "Bad schema expression ~a" schexp))))

(defun schexp-types (schexp)
  (cond ((sch-namep schexp)
	 (cond ((symbolp schexp)
		(if (var-decoratedp schexp)
		    (zsubst-alist 
		     (sloop for var in (sch-vars (undecorate-var schexp))
			    collect 
			    (cons var (decorate-var var (var-decoration schexp))))
		     (schema-types (undecorate-var schexp)))
		    (get-schema-types schexp)))
	       ((or (match schexp (DELTA x))
		    (match schexp (XI x)))
		(bind schexp (& x)
		      (let* ((vars (schexp-vars x))
			     (vars* (decorate-vars vars *prime-decoration*))
			     (types (schexp-types x)))
			(append types 
				(sloop for type in types
				       collect (zsubst-alist (pairlis vars vars*) type))))))))
	((consp schexp)
	 (cond ((match schexp (SCHEMA x) (schexp-types x)))
	       ((match schexp (NOT x)    (schexp-types x)))
	       ((match schexp (AND x y)  (union (schexp-types x) (schexp-types y))))
	       ((match schexp (OR x y)   (union (schexp-types x) (schexp-types y))))
	       ((match schexp (IMPLIES x y) (union (schexp-types x) (schexp-types y))))
	       ((match schexp (IFF x y)  (union (schexp-types x) (schexp-types y))))
	       ((match schexp (HIDE x y)
		       (let ((types (schexp-types x)))
			 (sloop for type in types
				when (null (intersection (all-vars type) y))
				collect type))))
	       ((match schexp (RESTRICT x y)
		       (let ((xvars (schexp-vars x))
			     (yvars (schexp-vars y)))
			 (schexp-types `(hide (and ,x ,y)
					      ,(set-difference xvars yvars))))))
	       ((match schexp (EXISTS x y)
		       (schexp-types `(hide (and ,x ,y) ,(schexp-vars x)))))
	       ((match schexp (PIPE x y)
		       (let ((qvars (piped-varnames (schexp-vars x)
						    (schexp-vars y))))
			 (union (schexp-types `(hide ,x ,(decorate-vars qvars 
									 *output-decoration*)))
				(schexp-types `(hide ,y ,(decorate-vars qvars 
									 *input-decoration*)))))))
	       ((match schexp (SEMI x y)
		       (let* ((xvars (schexp-vars x))
			      (yvars (schexp-vars y))
			      (qvars (sequenced-varnames xvars yvars)))
			 (union (schexp-types `(hide ,x ,(decorate-vars qvars *prime-decoration*)))
				(schexp-types `(hide ,y ,qvars))))))
	       ((match schexp (RENAME x y)
		       ;; y is a list (... (newvar-i oldvar-i) ...)
		       ;; It's possible for the user to rename two vars to the same one.
		       ;; Should this be an error?
		       (let ((types (schexp-types x))
			     (alist (sloop for pair in y collect 
					   (cons (cadr pair) (car pair)))))
			 (sloop for type in types
				collect (zsubst-alist alist type)))))
	       ((match schexp (PRE x)
		       (let ((vars (schexp-vars x)))
			 (schexp-types
			  `(hide ,x ,(append (vars-decorated-with vars *output-decoration*)
					     (vars-decorated-with vars *prime-decoration*)))))))
	       (t (error "Bad schema expression ~a" schexp))))
	(t (error "Bad schema expression ~a" schexp))))  

(defun schexp-axiom (schexp)
  (cond ((and (consp schexp) (eq (car schexp) 'schema))
	 (schexp-axiom (cadr schexp)))
        ((sch-namep schexp)
	 (cond ((symbolp schexp)
		(if (var-decoratedp schexp)
		    (let ((vars (var-names (sch-vars (undecorate-var schexp)))))
		      (zsubst-alist (pairlis vars (decorate-vars vars (var-decoration schexp)))
				    (schema-axiom (undecorate-var schexp))))
		    (schema-axiom schexp)))
	       ((match schexp (DELTA x)
		       (let* ((vars (var-names (schexp-vars x)))
			      (vars* (decorate-vars vars *prime-decoration*))
			      (axiom (schexp-axiom x)))
			 `(and ,axiom
			       ,(zsubst-alist (pairlis vars vars*) axiom)))))
	       ((match schexp (XI x)
		       (let* ((vars (var-names (schexp-vars x)))
			      (vars* (decorate-vars vars *prime-decoration*))
			      (axiom (schexp-axiom x)))
			 `(and ,axiom
			       ,(zsubst-alist (pairlis vars vars*) axiom)
			       ,@(sloop for var in vars
					as var* in vars*
					collect `(equal ,var ,var*))))))))
	((consp schexp)
	 (cond ((match schexp (SCHEMA x)
		       (schexp-axiom x)))
	       ((match schexp (NOT x) `(not ,(schexp-axiom x))))
	       ((match schexp (AND x y)
		       `(and ,(schexp-axiom x) ,(schexp-axiom y))))
	       ((match schexp (OR x y)
		       `(or ,(schexp-axiom x) ,(schexp-axiom y))))
	       ((match schexp (IMPLIES x y)
		       `(implies ,(schexp-axiom x) ,(schexp-axiom y))))
	       ((match schexp (IFF x y)
		       `(iff ,(schexp-axiom x) ,(schexp-axiom y))))
	       ((match schexp (HIDE x y)
		       (cond ((consp y)
			      `(exists ,y ,(schexp-axiom x)))
			     (t (error "Empty hiding ~a" schexp)))))
	       ((match schexp (RESTRICT x y)
		       (let ((xvars (var-names (schexp-vars x)))
			     (yvars (var-names (schexp-vars y))))
			 (schexp-axiom `(hide (and ,x ,y)
					      ,(set-difference xvars yvars))))))
	       ((match schexp (EXISTS x y)
		       (schexp-axiom `(hide (and ,x ,y) ,(var-names (schexp-vars x))))))
	       ((match schexp (PIPE x y)
		       (let ((qvars (piped-varnames (var-names (schexp-vars x))
						    (var-names (schexp-vars y)))))
			 (if (consp qvars)
			     (let ((x-subst (sloop for var in qvars
						   collect (cons (decorate-var var *output-decoration*)
								 var)))
				   (y-subst (sloop for var in qvars
						   collect (cons (decorate-var var *input-decoration*)
								 var))))
			       `(exists ,qvars
					(and ,(zsubst-alist x-subst (schexp-axiom x))
					     ,(zsubst-alist y-subst (schexp-axiom y)))))
			     `(and ,(schexp-axiom x) ,(schexp-axiom y))))))
	       ((match schexp (SEMI x y)
		       ;; What if there are no sequentially composed variables?
		       ;; Currently we let it pass without comment.
		       (let* ((xvars (var-names (schexp-vars x)))
			      (yvars (var-names (schexp-vars y)))
			      (vars (sequenced-varnames xvars yvars)))
			 (if (consp vars)
			     (let* ((qvars (sloop for var in vars
						  collect (make-new-var var 0 (append xvars yvars))))
				    (x-subst (sloop for var in vars
						    as qvar in qvars
						    collect (cons (decorate-var var *prime-decoration*)
								  qvar)))
				    (y-subst (sloop for var in vars
						    as qvar in qvars
						    collect (cons var qvar))))
			       `(exists ,qvars
					(and ,(zsubst-alist x-subst (schexp-axiom x))
					     ,(zsubst-alist y-subst (schexp-axiom y)))))
			     `(and ,(schexp-axiom x) ,(schexp-axiom y))))))
	       ((match schexp (RENAME x y)
		       ;; y is a list (... (newvar-i oldvar-i) ...)
		       ;; It's possible for the user to rename two vars to the same one.
		       ;; Should this be an error?
		       (cond ((consp y)
			      (zsubst-alist (sloop for pair in y collect 
						   (cons (cadr pair) (car pair)))
					    (schexp-axiom x)))
			     (t (error "Empty renaming ~a" schexp)))))
	       ((match schexp (PRE x)
		       (let ((vars (var-names (schexp-vars x))))
			 (schexp-axiom
			  `(hide ,x ,(append (vars-decorated-with vars *output-decoration*)
					     (vars-decorated-with vars *prime-decoration*)))))))
	       (t (error "Bad schema expression ~a" schexp))))
	(t (error "Bad schema expression ~a" schexp))))

(defun ztype-check (form type)
  ;; Form is the expression to be type-checked against type.
  ;; Type is of the forms below:
  (cond ((symbolp type)
	 (if (member type '(z integer integerp))
	     `(integerp ,form)
	     `(,type ,form)))
	((consp type)
	 (cond ((match type (pfun x y)
		       `(and (listp ,form)
			     (every (function (lambda (x) ,(ztype-check `(car x) x))) ,form)
			     (every (function (lambda (y) ,(ztype-check `(cdr y) y))) ,form)))) 
	       ((match type (tfun x y)
		       `(and (listp ,form)
			     (every (function (lambda (x) ,(ztype-check `(car x) x))) ,form) 
			     (every (function (lambda (x) ,(ztype-check `(cdr y) y))) ,form))))
	       ((match type (cross . l)
		       `(and
			 ,@(sloop for i from 0 to (- (length l) 1)
				  collect (ztype-check `(nth ,i ,form) (nth i l))))))
	       ((member (car type) '(powerset p1 f f1))
		`(every (function (lambda (x) ,(ztype-check 'x (cadr type)))) ,form)) 
	       ((match type (seq x)
		       `(every (function (lambda (x) ,(ztype-check 'x x))) ,form))) 
	       (t (error "Bad type specification ~a" type))))
	(t (error "Bad type specification ~a" type))))

(defvar *function-types* '(rel pfun tfun pinj inj psurj surj bij ffun finj))

(defvar *set-ops*
  (append '(powerset domain reange dres rres nrres ndres oplus image plus star
		     union intersection set* p1 f f1 bag hide id inv ) 
	  *function-types*))

;; (defun ztype-set-check (form symtab)
;;   ;; Form is the expression to be type-checked to see if its a set.
;;   ;; Ultimately needed to solve overloading problem on EQUAL.
;;   (cond ((symbolp form) (setvar-in-context? form context))
;; 	((member (car x) *set-ops) t)
;; 	((assoc form symtab)
;; 	 (let ((type (cadr (assoc form symtab))))
;; 	   nil))
;; 	(t nil)))

(defun var-names (l) (mapcar #'car l))

(defun sch-vars (schema-name)
  ;; Schema-name is an undecorated symbol.
  (let ((s (cdr (assoc schema-name *schemas*))))
    (if s (schema-vars s) 
	(error "Undefined schema name ~a:" schema-name))))

(defun sch-exists (schema-name)
  ;; Schema-name is an undecorated symbol.
  (let ((s (cdr (assoc schema-name *schemas*))))
    (if s (schema-exists s) 
	(error "Undefined schema name ~a:" schema-name))))

(defun sch-fn-vars (schema-name)
  ;; Schema-name is an undecorated symbol.
  (let ((event (symbol-function (ztype-fn schema-name))))
    (cond ((and (consp event) (eq (nth 0 event) 'defun))
	   (nth 2 event))
	  (t (error "Bad schema name ~a:" schema-name)))))

;(defun schema-axiom (schema-name)
;  ;; Schema-name is an undecorated symbol.
;  `(,(axiom-fn schema-name) ,@(var-names (sch-vars schema-name))))

(defun schema-axiom (schema-name)
  ;; Schema-name is an undecorated symbol.
  (schema-preds (cdr (assoc schema-name *schemas*))))


(defun get-schema-types (schema-name)
  (conjuncts (nth 3 (symbol-function (ztype-fn schema-name)))))

(defun piped-varnames (xvars yvars)
  (sloop for x in xvars
    when (and (output-var-p x)
	      (some (function 
			  (lambda (y) 
			    (equal y (decorate-var (undecorate-var x)
						    *input-decoration*))))
			 yvars))
    collect (undecorate-var x)))
  
(defun unpiped-vars (xvars yvars)
  (append 
   (sloop for x in xvars
     when (or (not (output-var-p x))
	      (not (some (function 
			  (lambda (y) 
			    (equal y (decorate-var (undecorate-var x)
						    *input-decoration*))))
			 yvars)))
     collect x)
   (sloop for y in yvars
     when (or (not (input-var-p y))
	      (not (some (function
			  (lambda (x)
			    (equal x (decorate-var (undecorate-var y)
						    *output-decoration*))))
			 xvars)))
     collect y)))

(defun sequenced-varnames (xvars yvars)
  (sloop for x in xvars
    when (and (primed-varp x)
	      (some (function 
		     (lambda (y) (equal y (undecorate-var x))))
		    yvars))
    collect (undecorate-var x)))
  
(defun unsequenced-vars (xvars yvars)
  (append 
   (sloop for x in xvars
     when (or (not (primed-varp x))
	      (not (some (function 
			  (lambda (y) (equal y (undecorate-var x))))
			 yvars)))
     collect x)
   (sloop for y in yvars
     when (not (some (function
		      (lambda (x)
			(equal x (decorate-var y *prime-decoration*))))
		     xvars))
     collect y)))

; ------------------------------------------------------------
; DEF-SCHEMA-BOX
;
; IN ACL2:
; We create several functions for schema, S.
;  S-type (vars) -> boolean: checks type-correctness of vars of S.
;  S-axiom (vars) -> boolean: checks that vars satisfy predicates of S.
;  S (state) -> boolean: ensures vars are bound in state and of the
;    right type.  Then checks that S-axiom holds.

;------------------------------------------------------------

; Create for each schema a structure.  From that structure we will
; derive a sequence of events (if working in ACL2) or Lisp defuns
; (if generating code or performing various adhoc checks).

; See the schema structure definition in structures.lisp.

; IN LISP:
; Wehn we generate functions we create several functions for schema, S.
;  S-type (vars) -> boolean: checks type-correctness of vars of S.
;  S-axiom (vars) -> boolean: checks that vars satisfy predicates of S.
;  S (input) -> output: ensures vars are bound in state and then "executes".
;
; (progn (defn schema-type (v1 ... vn) properly-typed(v1 .. vn))
;        (defn schema-axiom (st) (bind (v1 ... vn) st 
;                                      (-> (schema-type v1 .. vn)
;                                          (and . AXIOM-PART)))
;        (defn schema-exec (?1 ... ?n) (mv !rhs1 ... !rhs2))

(defun def-schema-box (schema-name decl-part where)
  (let* ((name (make-schema-name schema-name))
	 (locals (schema-local-vars decl-part))
	 (vars (schema-box-vars decl-part))
	 (primes (remove-if-not #'primed-var-p vars))
         ;; Vars is a list of (name . type)
	 (axioms (andlist (append (decl-axioms decl-part) where)))
	 (functions (new-schema-functions decl-part))
	 (schemas (decl-schemas decl-part)) 
	 (signature (vars-signature vars)))
    (setq *schemas*
	  (cons (cons name
		      (make-schema :name name
				   :kind 'schema
				   :decls decl-part
				   :locals locals
				   :vars vars
				   :primes primes
				   :schemas schemas
				   :signature signature
				   :preds axioms
				   :functions functions
				   :input  (retain vars (function input-var-p))
				   :output (retain vars (function output-var-p))
				   :exists (extract-existentials axioms)
					; :exists exists
					; :exec ...
				   ))
		*schemas*))))

(defun schema-eval (form &optional (s *current-schema*))
  (let* ((s (if s (schema-p s) (schema-p (car *schemas*))))
	 (l (var-names (cdr (schema-signature s)))))
    (progv l (mapcar (function (lambda (x) (getv x s))) l)
	   (eval form))))

(defun extract-existentials (axioms)
  (cond ((atom axioms) nil)
	((eq (car axioms) 'exists)
	 (append (extract-variables (cadr axioms))
		 (extract-existentials (caddr (cdr axioms)))))
	(t (sloop for pred in (cdr axioms)
		  appending (extract-existentials pred)))))

(defun def-executable-schemas (l) (mapcar (f-l (x) (defun-schema-fn (cdr x))) l))

(defun mentions (pred vs)
  (cond ((null vs) nil)
	((symbolp pred) (member pred vs))
	(t (some (function (lambda (x) (mentions x vs))) (cdr pred)))))

(defun defun-schema-fn (s)
  "Create executable lisp from schemas that describe operations
The resulting function depends on SELF (computed by looking up the 
schema s in the *schema* list) to provide the context for
variable lookup."
  (let* ((name   (schema-name s))
	 (input  (schema-input s))
	 (ivars  (var-names input))
	 (output (schema-output s))
	 (ovars  (var-names output))
	 (primes (schema-primes s))
	 (notprimes (set-difference! (schema-vars s)
				     (union! input (union! output primes))))
	 (pvars  (var-names primes))
	 (svars  (set-difference! (schema-vars s) (union! input (union! output primes))))
	 (axioms (schema-preds s))
	 (functions (schema-functions s))
	 (fnname (executable-fn name))
	 (invars  (append ivars (var-names notprimes)))
	 (outvars (append ovars (var-names primes))))
    (declare (ignore invars svars))
    (mapc (f-l (f)
	       (let ((fn (car f)))
		 (eval `(defun ,fn (x) (cdr (assoc x (getv ',fn)))))))
	  functions)
    ;; NOTE: (and) evals to T
    (eval `(defun ,fnname (,@ivars)
	     (let ((*current-schema* (cdr (assoc ',name *schemas*)))
		   ,@(sloop for v in (var-names notprimes)
			    collect (list v `(getv ',v ',name))))
	       ,@(sloop for v in ivars
			collect `(setv ',v ,v))
	       ,@(sloop for vartype in input
			collect `(if (not ,(ztype-check (car vartype) (cdr vartype)))
				     (warning "~%~s given input arg of wrong type: ~s"
					      ',fnname ',(car vartype))))
	       ,@(sloop for vartype in notprimes
			collect `(if (not ,(ztype-check (car vartype) (cdr vartype)))
				     (warning "~%~s given state arg of wrong type: ~s"
					      ',fnname ',(car vartype))))
	       ,@(sloop for pred in (conjuncts axioms)
			when (not (mentions pred outvars))
			collect `(if (not ,pred)
				     (warning "~%~s has a failed precondition:~% ~s"
					      ',fnname
					      ',pred)))
	       (let* ,(collect-computable-functions (append output primes)
						    (conjuncts axioms)
						    (append input notprimes)
						    name)
		 (declare (special ,@ovars ,@pvars))
		 ,@(if ovars `((mapc (f-l (v) (setv v (eval v) ',name)) ',ovars)) nil)
		 ,@(if pvars `((mapc (f-l (v) (setv v (eval v) ',name)) ',pvars)) nil)
		 ,@(if output
		       (sloop for vartype in output
			      collect
			      `(if (not (and ,(ztype-check (car vartype)
							   (cdr vartype))))
				   (warning "~%~s produces output values of wrong type: ~s"
					    ',fnname
					    ',(car vartype))))
		       nil)
		 ,@(if output
		       (sloop for vartype in primes
			      collect
			      `(if (not ,(ztype-check (car vartype)
						     (cdr vartype)))
				   (warning "~%~s produces state output values of wrong type: ~s"
					    ',fnname
					    ',(car vartype)))))
		 ,@(sloop for pred in (conjuncts axioms)
			  when (mentions pred outvars)
			  collect `(if (not ,pred)
				       (warning "~%~s has a failed postcondition:~% ~s"
						',fnname
						',pred)))
		 (values ,@ovars)))))
    (setf (get fnname 'schema) s)
    (setf (schema-exec s) (cons fnname output))
    fnname)) 

(defun collect-computable-functions (outputs preds input context)
  ;; Returns a form suitable for inclusion as arglist in a let*
  ;;  ((var form)*) 
  ;; Context is a schema name.
  (let* ((poss (insert-choose-as-necessary (sortpreds preds outputs input) outputs nil context))
	 fun body)
    ;; POSS is a list of the form (((id . type) depends form)*)
    (sloop for cand in poss
	   when (and (not (assoc (caar cand) body))
		     (setq fun (generate-function cand context)))
	   do (setq body (append body (list fun))))
    body))
	       
(defun generate-function (cand context)
  ;; Cases supported
  ;;  ((Y . INT) dep (EQUAL Z (FOB Y)))
  ;;  ((X . INT) dep (FOD Y S X))        -- turns into choose with constraint.
  ;;  ((V . INT) NIL (CHOOSE V INT CONTEXT))
  ;; When calling a 'schema' we need to save the input/output mapping so that
  ;; we can provide the same output if called with prviously used inputs.
  ;; RETURN: (id function)
  (let ((id (caar cand))
	(type (cdar cand))
	(pred (caddr cand)))
    (cond ((and (or (equal (car pred) 'equal) (equal (car pred) 'equal*))
		(equal id (cadr pred)))
	   (list id (caddr pred)))
	  ((and (or (equal (car pred) 'equal) (equal (car pred) 'equal*))
		(equal id (caddr pred)))
	   (list id (cadr pred)))
	  ((equal (car pred) 'choose) pred)
	  (t (list id `(choose ',id ',type  ',context (function (lambda (,id) ,pred))))))))

(defun var-ids (l) (mapcar #'car l))

(defun remove-var (v l) (remove v l :test #'equal))

(defun make-choose-tuple (var context)
  (list var nil `(choose ,(car var) ,(cdr var) ',context)))

(defun insert-choose-as-necessary (l output input context)
  ;; For any missing dependent before a pred, insert a choose.
  ;; Insert a choose for the leftovers.
  (if (null l)
      (if (null output)
	  nil
	  (mapcar (f-l (o) (make-choose-tuple o context)) output))
      (let ((var (caar l))
	    (dep (cadar l))
	    new)
	(if (not (subsetp dep input :test #'equal))
	    (progn
	      (setq new (append (mapcar (f-l (o) (make-choose-tuple o context))
					(set-difference! dep input))))
	      (setq output (set-difference! output dep))
	      (setq input (union! input dep))))
	(append new
		(cons (car l)
		      (insert-choose-as-necessary (cdr l)
						  (remove-var var output)
						  (adjoin var dep :test #'equal)
						  context))))))

(defun extract-outputs (l outputs)
  (cond ((not (consp l))
	 (if (and (symbolp l) (assoc l outputs))
	     (list (assoc l outputs))
	     nil))
	(t (append (extract-outputs (car l) outputs)
		   (extract-outputs (cdr l) outputs)))))

(defun insert-state-lookup (l outputs)
  (cond ((not (consp l))
	 (cond ((not (symbolp l)) l)
	       ((assoc l outputs) l)
	       (t `(getv ',l))))
	(t (cons (car l)
		 (mapcar (f-l (x) (insert-state-lookup x outputs)) (cdr l))))))

(defun make-tuples (cand output input)
  (let ((inputs (extract-outputs cand output))
	;(form   (insert-state-lookup cand (append input output)))
	(form   cand))
    (cond ((null inputs) nil)
	  ((null (cdr inputs)) (list (list (car inputs) nil form)))
	  (t (sloop for x in inputs
		    collect (list x (remove-var x inputs) form))))))
 
(defun var-and-equal (a b)
  (if (equal (car a) (car b))
      (and (consp (caddr a))
	   (equal (car (caddr a)) 'equal)
	   (equal (cadr (caddr a)) (caar a)))
      nil))

(defun depends-on (l done)
  ;; Elements of l that depend only on elements of done.
  (sort 
   (sloop for x in l
	  when (subsetp (cadr x) done :test #'equal)
	  collect x)
   #'var-and-equal))

(defun sortpreds (l output input) 
  (sortpreds2 (mapcon (f-l (cdr) (make-tuples (car cdr) output input)) l) nil))

(defun sortpreds2 (l done) 
  (if (null l)
      nil
      (let ((dep (depends-on l done)))
	(if (null dep)
	    l
	    (append dep (sortpreds2 (set-difference! l dep)
				    (union! done (mapcar #'car dep))))))))  

(defun decl-axioms (decl-part) 
  (sloop for decl in decl-part
	 when (and (consp decl)
		   (or (equal (car decl) 'schema)
		       (equal (car decl) 'delta)
		       (equal (car decl) 'xi)
		       (and (equal (car decl) 'basic-decl) (null (cadr decl)))))
	 collect (schexp-axiom (if (equal (car decl) 'basic-decl)
				   (caddr decl)
				   decl))))

(defun schema-local-vars (decl-part) 
  ;; Extract local variable (name . type) pairs from the declaration part.
  (no-duplicates
   (sloop for decl in decl-part
     append (cond ((match decl (basic-decl nil . schemas)) nil)
		  ((match decl (basic-decl type . vars)
			  (mapcar (function (lambda (x) (cons x type))) vars)))
		  (t nil)))))

(defun schema-box-vars (decl-part) 
  ;; Extract variable (name . type) pairs from the declaration part.
  (no-duplicates
   (sloop for decl in decl-part
     append (cond ((match decl (basic-decl nil . schemas)
			  (mapcan (function (lambda (x) (schexp-vars x))) schemas)))
		  ((match decl (basic-decl type . vars)
			  (mapcar (function (lambda (x) (cons x type))) vars)))
		  ((match decl (SCHEMA x))
		   (schexp-vars (cadr decl)))
		  ((match decl (DELTA x))
		   (append (schexp-vars (cadr decl))
			   (decorate-vars (schexp-vars (cadr decl))
					  *prime-decoration*)))
		  ((match decl (XI x))
		   (append (schexp-vars (cadr decl))
			   (decorate-vars (schexp-vars (cadr decl))
					  *prime-decoration*)))
		  (t (error "Bad decl ~a:" decl))))))

(defun ztype-function (type)
  (cond ((symbolp type) type)
	((consp type)
	 (cond ((match type (pfun x y)
		       (pfun-typename (ztype-function x) 
				      (ztype-function y))))
	       ((match type (tfun x y)
		       (pfun-typename (ztype-function x) 
				      (ztype-function y))))
	       ((match type (cross . l)
		       (cross-typename (mapcar (f-l (x) (ztype-function x)) l))))
	       ((member (car type) '(powerset p1 f f1))
		(powerset-typename (ztype-function (cadr type))))
	       ((match type (seq x)
		       (seq-typename (ztype-function x))))
	       (t (error "Bad type specification ~a" type))))
	(t (error "Bad type specification ~a" type))))

;; STEP updates the global state by setting all variables with primed versions
;; to the primed value and CLEARING THE PRIMED VALUES.

(defun zstep ()
  (mapcar #'update-schema *schemas*))

(defmacro test (&rest calls)
  (if (null calls)
      `(random-test)
      `(test2 ',calls t)))
  
(defun random-test () (format t "~%Random execution of calls not implemented~%"))

(defun test2 (calls &optional print-state)
  (if (null calls)
      nil
      (let* ((form (car calls))
	     (fn (car form))
	     (schema (get fn 'schema)))
	(if print-state (format t "~%Evaling ~s" form))
	(eval form)
	(if print-state (print-bindings schema))
	(zstep)
	(if print-state (format t "~%Step" form))
	(test2 (cdr calls) print-state))))

(defun update-schema (pair)
  (if (schema-next (cdr pair))
      (let ((new (schema-p (schema-next (cdr pair)))))
	;; Can we rely on bindings being in the same order?
	;; I think we can, but I choose not to.
	(if (schema-updated new)
	    (let ((newbindings (schema-bindings new)))
	      (setf (schema-bindings (cdr pair))
		    (mapcar (function (lambda (x)
					(cons x
					      (cdr (assoc (decorate-var x
									*prime-decoration*)
							  newbindings)))))
			    (var-names (schema-locals (cdr pair)))))
	      (setf (schema-bindings new) nil)
	      (setf (schema-updated new) nil))))))	    

; DECL-TYPE-GENERATORS creates forms to introduce new types
; encountered in the decl-part of a schema box.

(defun decl-type-generators (types)
  (sloop for decl in types
    append (new-type-generators decl)))

(defun new-schema-types (decl-part)
  (sloop for decl in decl-part
    append (cond ((match decl (basic-decl nil . schemas)) nil)
		 ((match decl (basic-decl type . vars) (list type)))
		 ((match decl (schema x)) nil)
		 ((match decl (DELTA x)) nil)
		 ((match decl (XI x)) nil)
		 (t (error "Bad decl ~a:" decl)))))

(defun new-schema-functions (decl-part)
  ;; Two ways this can wrork.  In this case
  ;; we are looking for pfun, fun, pinj, inj, psurg, surj, or ffun
  ;; which will just look up their associated value in the alist 
  ;; representing their value.
  (sloop for decl in decl-part
    append (cond ((match decl (basic-decl nil . schemas)) nil)
		 ((match decl (basic-decl type . vars))
		  (let ((type (cadr decl)))
		    (if (and (consp type)
			     (member (car type) *function-types*))
			(mapcar (f-l (x) (cons x type)) (cddr decl))
			nil)))
		 ((match decl (schema x)) nil)
		 ((match decl (DELTA x)) nil)
		 ((match decl (XI x)) nil)
		 (t (error "Bad decl ~a:" decl)))))

(defun schema-functions-in-preds (functions preds)
  ;; functions = ((basic-decl type . y)*)
  ;; Where type is one of the relation constructors (e.g. pfun ...)
  (let* ((name-types (mapcon (f-l (x)
				  (mapcar (f-l (y) (list y (cadr x)))
					  (cddr x)))
			     functions)))
    (if name-types
	(find-function-defns name-types preds)
	nil)))

(defun find-function-defns (name-types preds)
  (let (function-part)
    (sloop for term in preds
	   when (match term (every x type form)
		       (setq function-part (check-function-match form name-types)))
	   collect (prog1 function-part (setq function-part nil)))))

(defun check-function-match (form name-types)
  ;; Does form define a part of an element of name-types in an equality.
  (if (and (consp form) (eq (car form) 'equal))
      (let ((left (cadr form))
	    (right (caddr form)))
	(cond ((and (consp left) (assoc (car left) name-types))
	       (list (car left) right))
	      ((and (consp right) (assoc (car right) name-types))
	       (list (car right) left))
	      (t nil)))
      nil))

(defun decl-schemas (decl-part)
  "This function needs to create schemas if they don't exist.
In particular if S' does not exist and we hit (DELTA S) or (XI S)
we need to create the primed version of S."
  (sloop for decl in decl-part
    append (cond ((match decl (schema x)
			 (list (defined-schema x))))
		 ((match decl (DELTA x)
			 (list (defined-schema x)
			       (defined-schema
				 (decorate-var x *prime-decoration*)))))
		 ((match decl (XI x)
			 (list (defined-schema x)
			       (defined-schema
				 (decorate-var x *prime-decoration*)))))
		 ((match decl (basic-decl nil . vars) vars))
		 ((match decl (basic-decl type . vars)) nil)
		 (t (warning "Bad decl ~a:" decl)))))

; NEW-TYPE-GENERATORS uses the conventions for naming 
; type recognizers, and generates Nqthm forms that define
; newly encountered types within the type expressions TYPE.

(defun new-type-generators (type)
  (cond ((symbolp type) nil)
	((consp type)
	 (cond ((match type (pfun x y)
		       (let ((fn (pfun-typename (ztype-function x)
						(ztype-function y))))
			 (append (new-type-generators x)
				 (new-type-generators `(powerset ,x))
				 (new-type-generators y)
				 (new-type-generators `(powerset ,y))
				 ;; `((defalist ,fn (a) 
				 ;; ,(ztype-function x)
				 ;; ,(ztype-function `(powerset ,x))
				 ;; ,(ztype-function y)
				 ;; ,(ztype-function `(powerset ,y))))
				 `((defalist ,fn (l) (,(ztype-function x) . ,(ztype-function y))))
				 ))))
	       ((match type (cross . l)
		       (error "Can't handle Cross type specifications yet ~a"
			      type)))
	       ((member (car type) '(powerset p1 f f1))
		(let* ((x (cadr type))
		       (fn (powerset-typename (ztype-function x))))
		  (append (new-type-generators x)
			  `((defset ,fn (l) ,(ztype-function x))))))
	       ((match type (seq x)
		       (let ((fn (seq-typename (ztype-function x))))
			 (append (new-type-generators x)
				 `((deflist ,fn (l) ,(ztype-function x)))))))
	       (t (error "Bad type specification ~a" type))))
	(t (error "Bad type specification ~a" type))))

(defvar *chronology* nil)

(defun set-theory-defined (fn)
  (member (pack fn '-union) *chronology*))

(defun seq-theory-defined (fn)
  (member (pack fn '-nthcdr) *chronology*))

(defun bag-theory-defined (fn)
  (member (pack fn '-bagint) *chronology*))

(defun def-generic (name args schexp)
  (let ((vars (schexp-vars schexp))
	;; existential vars
	(existentials (schexp-existentials schexp))
	;; pred is a single expression ???
	(pred (schexp-axiom schexp)))
    (setq *generics* (cons (cons name
				 (make-schema :name name
					      :kind 'generic
					      :gen-formals args
					      :vars vars
					      :exists existentials
					      :preds pred))
			   *generics*))))

; ------------------------------------------------------------
; DEF-GENERIC-BOX
;
; We create a macro for generic, (G args).
; G-macro (args) > { S-type (vars) -> boolean),
;                    S-axiom (vars) -> boolean,
;                    S-fun (state) -> boolean }

;------------------------------------------------------------

(defmacro def-generic-box (generic-header decl-part &key where)
  (let ((name (if (symbolp generic-header) generic-header (car generic-header)))
	(args (if (symbolp generic-header) nil (cdr generic-header))))
  `(defmacro ,name ,args
    ,(substitute-commaized-args 
      args
      `(def-schema-box ,(new-schema-name name)
	 ',decl-part ',where)))))

(defun substitute-commaized-args (args form)
  (sloop for arg in args
	 do (setq form (substitute (commaize arg) arg form)))
  form)

(defun commaize (x) (pack '|,| x))


; ------------------------------------------------------------
; DEF-AXIOM : (def-axiom decls axioms)
; ------------------------------------------------------------

; Example
; 
;   (def-axiom ((basic-decl (f person) borrower staff)
;               (basic-decl (pfun person book) owner)
; 	       ((< x y)))
;   =>
;   (defstub x-value () t)
;   (defstub y-value () t)
;   (defaxiom axiom-3 (< (x-value) (y-value)))
;   ->
;   (def-variable x integer :pred (< x y))
;   (def-variable y integer :pred (< x y))

;   (def-axiom ((var (pfun integer integer) square))
; 	       ((forall n integer (equal (square n) (* n n)))))
;   =>
;   (encapsulate ((square (integer) t))
;                (local (defun square (n) (* n n)))
;                (defaxiom axiom-1 (implies (integerp n) (equal (square n) (* n n))))
;                (defaxiom axiom-2 (implies (integerp n) (integerp (square n)))))
;   ->
;   (defun square (n) (if (integerp n) (* n n) 'bottom))

;   (def-axiom ((var (pfun integer integer) square))
; 	       nil)
;   =>
;   (defstub square (integer) t)
;   (defaxiom axiom-2 (implies (integerp n) (integerp (square n))))
;   ->
;   (defun square (n) (choose n 'x 'integer nil))

(defun update-globals (vars preds)
  (setq *globals*
	(append
	 (mapcar (f-l (l)
		      (let* ((preds (relevant-preds (car l) preds))
			     (restriction (if preds (cons 'and preds) nil)))
			(cons (car l) (make-zglobal :name (car l)
						    :type (cdr l)
						    :pred restriction))))
		 vars)
	 *globals*)))

(defun relevant-preds (var preds)
  (sloop for p in preds
	 when (in-p var p)
	 collect p))

(defun def-executable-vars (globals) (mapc (function defvar-global) globals))
(defun defvar-global (x)
  (proclaim (list 'special (car x)))
  (let ((g (cdr x)))
    (if (zglobal-value g)
	(set (zglobal-name g) (zglobal-value g))
	(set (zglobal-name g) (choose (zglobal-name g) (zglobal-type g) nil (zglobal-pred g))))))

(defun def-executable-constants (constants) (mapc (function defvar-constant) constants))
(defun defvar-constant (pair)
  (proclaim (list 'special (car pair)))
  (set (car pair) (car pair)))

;; (choose (car pair) (cdr pair))))

;; (add-type name def base function first)
;; (add-type x nil x (function true) x)) l))

(defun def-axiom (decl-part &optional where)
  (let* ((vars  (schema-box-vars decl-part))
	 (functions (new-schema-functions decl-part))
	 (preds (andlist (append (decl-axioms decl-part) where)))
	 (defns (schema-functions-in-preds functions preds))
	 (signature (vars-signature vars)))
    (declare (ignore signature))
    (update-globals vars (append (decl-axioms decl-part) where))
    (setq *axioms*
	  (cons (make-schema :name nil
			     :vars vars
			     :functions functions
			     :preds preds
			     :defns defns
			     )
		*axioms*))))

(defmacro defun-axiom-fn (a)
  (let* ((s (eval a))
	 ;(name 'axiom)
	 ;(types  (schema-types s))
	 (vars   (schema-vars s))
	 (preds  (schema-preds s)))
    `(defun ,(axiom-fn (car s)) ,vars
       (if ,(conjunct (make-global-assertions vars))
	   ,preds
	   nil))))

; REMARK the call to schexp-types below returns
;        a flattened set of conjuncts. Whenever x is
;        an undecorated symbol, we could produce a call to
;        the type fn.

(defun interpret-zterm (term)
  (cond ((atom term) term)
	((fquotep term) term)
	((consp term)
	 (cond ((match term (schema x)
		       (andlist (cons (schexp-axiom x)
				      (schexp-types x)))))
	       (t (cons-term (ffn-symb term)
			     (sloop for arg in (ffn-args term)
				    collect (interpret-zterm arg))))))))

; ------------------------------------------------------------
; Variable Decorations
; ------------------------------------------------------------

(defun output-varp (var)
  (let ((d (symbol-name (var-decoration var))))
    (and (< 0 (length d))
	 (equal (char d (1- (length d))) *output-decoration-char*))))

(defun input-varp (var)
  (let ((d (symbol-name (var-decoration var))))
    (and (< 0 (length d))
	 (equal (char d (1- (length d))) *input-decoration-char*))))

(defun primed-varp (var)
  (let ((d (symbol-name (var-decoration var))))
    (and (< 0 (length d))
	 (equal (char d (1- (length d))) *prime-decoration-char*))))
		    
(defun primed-var-p (x) 
  ;; X may be a symbol, but should really be a (var.type) pair.
  (if (symbolp x)
      (primed-varp x)
      (if (consp x)
	  (primed-varp (car x))
	  nil)))

(defun input-var-p (x) 
  ;; X may be a symbol, but should really be a (var.type) pair.
  (if (symbolp x)
      (input-varp x)
      (if (consp x)
	  (input-varp (car x))
	  nil)))

(defun output-var-p (x) 
  ;; X may be a symbol, but should really be a (var.type) pair.
  (if (symbolp x)
      (output-varp x)
      (if (consp x)
	  (output-varp (car x))
	  nil)))

(defun decorate-var (var decoration)
  ;; BOGUS overloading.  Handles symbols and (name.type).
  (if (consp var)
      (cons (pack (car var) decoration) (cdr var))
      (pack var decoration)))

(defun decorate-vars (symbols decoration)
  (sloop for symbol in symbols collect 
    (decorate-var symbol decoration)))

(defun var-decoration (var)
  (let ((s (symbol-name (if (consp var) (car var) var))))
    (pack* (reverse (sloop for i downfrom (1- (length s)) to 0
		     while (member (char s i) *decoration-chars*)
		     collect (char s i))))))

(defun remove-one-decoration (symbol)
  (if (var-decoratedp symbol)
      (let* ((name (symbol-name symbol))
	     (s (subseq name 0 (1- (length name)))))
	(intern s))
      symbol))

(defun var-decoratedp (symbol)
  (let ((s (symbol-name symbol)))
    (member (char s (1- (length s)))
	    *decoration-chars*)))

(defun vars-decorated-with (vars decoration)
  (sloop for var in vars
    when (and (var-decoratedp var)
	      (equal (var-decoration var) decoration))
    collect var))

(defun undecorate-var (symbol)
  (if (consp symbol)
      (cons (undecorate-var2 (car symbol)) (cdr symbol))
      (undecorate-var2 symbol)))

(defun undecorate-var2 (symbol)
  (let ((s (symbol-name symbol)))
    (pack* (sloop for i from 0 to (1- (length s))
	      while (not (member (char s i) *decoration-chars*))
	      collect (char s i)))))

; ------------------------------------------------------------
; Choose function
; 
; We want all of the variables in pred to have been instantiated
; in the calling fn.  The pred should have been backquoted in the calling
; environment.

(defvar choicemax 10)
(defvar *choices* nil)

(defun reset-choices () (setq *choices* nil))

(defun choose (var type context &optional pred (count 0))
  (let (value)
    (cond ((> count choicemax)
	   (format t "~%Exceeding bound for guessing a value for ~s of type ~s in context ~s~%~
                      Please input a value:" var type context)
	   (setq value (read))
	   (setv var value context)
	   value)
	  ((setq value (getv var context))
	   ;; If there is no predicate, leave it.
	   ;; If the current value satisfies the predicate, leave it.
	   (cond ((null pred) value)
		 ((progv (list var) (list value)
			 (eval pred))
		  value)
		 (t (setq value (next-choice var type value pred))
		    (setv var value context)
		    (choose var type context pred (+ count 1)))))
	  ;; If there is no curent value, pick one.
	  (t (setq value (first-choice var type pred))
	     (setv var value context)
	     (if (null pred)
		 value
		 (choose var type context pred (+ count 1)))))))

(defun first-choice (var type pred)
  (cond ((symbolp type)
	 (if (member type '(z integer integerp))
	     0
	     (let ((struct (cdr (assoc type *types*))))
	       (if (null struct)
		   (error (format nil "Reference to undefined typename, ~s" type))
		   (let* ((firstchoice (ztype-firstchoice struct))
			  (basetype (ztype-basetype struct))
			  ;(def (ztype-def struct))
			  )
		     (cond (firstchoice firstchoice)
			   (basetype (first-choice var basetype pred))
			   ;(def (first-choice var def pred))
			   (t (error (format nil "Ill-formed typedef for ~s" type)))))))))
	((consp type)
	 (cond ((match type (pfun x y)) nil)
	       ((match type (tfun x y)) nil)
	       ((match type (cross . l)
		       (sloop for i from 0 to (- (length l) 1)
			      collect (first-choice `(nth ,i ,var) (nth i l) pred))))
	       ((member (car type) '(powerset p1 f f1)) nil)
	       ((match type (seq x)) nil)
	       (t (error "Bad type specification ~a" type))))
	(t (error "Bad type specification ~a" type))))

(defun next-choice (var type last pred)
  (cond ((symbolp type)
	 (if (member type '(z integer integerp))
	     (if last (+ last 1) 0)
	     (let ((struct (cdr (assoc type *types*))))
	       (if (null struct)
		   (error (format nil "Reference to undefined typename, ~s" type))
		   (let* ((basetype (ztype-basetype struct))
			  ;(def (ztype-def struct))
			  )
		     (cond (basetype (next-choice var basetype last pred))
			   ;(def (next-choice var def last pred))
			   (t (error (format nil "Ill-formed typedef for ~s" type)))))))))
	((consp type)
	 (cond ((match type (pfun x y) 
		       (cons (cons
			      (next-choice `(caar ,var) x 
					   (if (null last)
					       (first-choice `(caar ,var) x pred)
					       (caar last))
					   pred)
				   (next-choice `(cdar ,var) y 
						(if (null last)
						    (first-choice `(caar ,var) y pred)
						    (cdar last))
						pred))
			     last)))
	       ((match type (tfun x y)
		       (cons (cons (next-choice `(caar ,var) x 
						(if (null last)
						    (first-choice `(caar ,var) x pred)
						    (caar last))
						pred)
				   (next-choice `(cdar ,var) y 
						(if (null last)
						    (first-choice `(caar ,var) y pred)
						    (cdar last))
						pred))
			     last)))
	       ((match type (cross . l)
		       (sloop for i from 0 to (- (length l) 1)
			      collect (next-choice `(nth ,i ,var)
						   (nth i l)
						   (if (null last)
						       (first-choice `(nth ,i ,var)
								     (nth i l) pred)
						       (nth i last))
						   pred))))
	       ((member (car type) '(powerset p1 f f1))
		(let ((x (cadr type)))
		  (cons (next-choice `(car ,var)
				     x
				     (if (null last)
					 (first-choice `(car ,var) x pred)
					 (car last))
				     pred)
			last)))
	       ((match type (seq x)
		       (cons (next-choice `(car ,var)
					  x
					  (if (null last)
					      (first-choice `(car ,var) x pred)
					      (car last))
					  pred)
			     last)))
	       (t (error "Bad type specification ~a" type))))
	(t (error "Bad type specification ~a" type))))
  
;; --------------------------------------------------------------
;; Loading ztc files


(defparameter *parser* "/usr/cli/projects/z/ez/parser/parse")

(defvar *testread* 
  (let ((*readtable* (copy-readtable nil)))
    (set-syntax-from-char #\' #\a)
    *readtable*))

(defvar a-very-rare-cons (cons nil nil))

(defun loadz (file)
  (let* ((fullname (namestring (probe-file file)))
	 (mirror (concatenate 'string fullname ".mirror"))
	 (parse  (format nil "~a ~a > ~a" *parser* fullname mirror)))
    (if (probe-file file)
	(multiple-value-bind (unix-status other) (si::system parse)
	  (declare (ignore other))
	  (if (equal unix-status 3)
	      (error "Z parse of input file failed.")
	      (with-open-file
	       (*standard-input* mirror :direction :input)
	       (let ((*readtable* *testread*))
		 (sloop for form = (read *standard-input* nil a-very-rare-cons nil)
		  until (eq form a-very-rare-cons)
		  do (eval (normalize form)))))))
	(format t "File ~s does not exist" file))))

(defun normalize (form &optional symtab)
  ;; (schematext ((basic-decl nil x))) => x (for now)
  (cond ((atom form) form)
	((consp (car form)) (cons (normalize (car form)) (normalize (cdr form))))
	((eq (car form) 'parens) (normalize (cadr form)))
	((eq (car form) 'equal)  (cons 'equal* (normalize (cdr form))))
	((eq (car form) 'forall) (cons 'every (normalize (cdr form))))
	((eq (car form) 'exists) (cons 'some  (normalize (cdr form))))
	((eq (car form) 'copy) (cons 'copy* (normalize (cdr form))))
	((match form (forall (schematext ((basic-decl type . x))) expr)
		`(every (function (lambda (,@x) ,(normalize expr))) ,(normalize type))))
	((match form (exists (schematext ((basic-decl type . x))) expr)
		`(some (function (lambda (,@x) ,(normalize expr))) ,(normalize type))))
	((match form (exists1 (schematext ((basic-decl type . x))) expr)
		`(and (some (function (lambda (,@x) ,(normalize expr)))
			    ,(normalize type))
		      (eq (length ,(normalize type)) 1))))
	((match form (set (schematext ((basic-decl nil . x))))
		(cons 'set* x)))
	((eq (car form) 'set) (cons 'set* (normalize (cdr form))))
	((match form (set-comp (schematext ((basic-decl nil . x))))
		(cons 'set* x)))
	((match form (set* (schematext ((basic-decl nil . x))))
		(cons 'set* x)))
	((match form (schematext ((basic-decl nil . x)))
		x))
	(t (cons (car form) (normalize (cdr form))))))

;; ----------------------------------------------------------------
;; DEAD CODE

(defun deconstruct (arg x name constants)
  (cond ((symbolp x)
	 (cond ((member x constants) `(equal ,arg ',x))
	       ((eq x name) `(,name ,arg))
	       ((assoc x *types*) `(,x ,arg))
	       (t (error (format nil "Bad type ref, ~a, in def-ident" x)))))
	(t `(and ,@(sloop for i from 0 to (- (length x) 1)
			  collect (deconstruct (list 'nth i arg)
					       (nth i x)
					       name
					       constants))))))




; ------------------------------------------------------------
; Changes:
;
; zmacros  -> zmacros1
;
;   1. Code reorganization.
;
; zmacros1 -> zmacros2
;
;   1. Fixed error in def-schema and def-schema-box macros.
;      These should generate distinct predicates about the
;      the types of variables, and about the variable axioms.
;      Bevier and Cohen. 12/4/94
;   2. Handle EXISTS schema operator. Bevier
;
; zmacros2 -> zmacros3
;
;   1. Added def-type macro. Bevier
;   2. Handle RESTRICT schema operator. Bevier
;
; zmacros3 -> zmacros4
;
;   1. Created ability to automatically generate type 
;      recognizers from the decl part of a schema box.
;      This feature uses defstructure macros in 
;      /home4/os/libs/defstructures.lisp. Bevier 12/27/94
;   2. Created def-ident macro. Bevier 12/27/94
;
; zmacros4 -> zlisp.lisp
;
;   1. Modify for Acl2.
;   2. iterate => sloop

; ------------------------------------------------------------

(setq x (make-schema
	 :name 'foo
	 :kind 'schema
	 :input  '((a! . integerp)(b! . integerp))
	 :output '((x! . integerp)(y! . integerp)(z! . integerp))
	 :preds '(and (equal x! (+ a! 1)) (equal y! (+ x! 10)) (< z! y!))))

(setq y (make-schema
	 :name 'foo
	 :kind 'schema
	 :vars  '((x! . integerp)(y! . integerp)(z! . integerp)(w! . integerp))
	 :input  '(a? b?)
	 :output '(x! y! z! w!)
	 :preds '(and (equal x! (+ a? 1)) (equal b? (+ w! 1)) (equal y! (+ x! 10)) (< z! y!))))

(def-free-type 'bool
  '((constant true)(constant false)
    (constructor Band (cross bool bool))
    (constructor Bor (cross bool bool))
    (constructor Bnot bool)))

(defun print-schema (s)
  (let ((name (schema-name s))
	(formals (schema-gen-formals s))
	(kind (schema-kind s))

	(vars (schema-vars s))
	(schemas (schema-schemas s))
	(functions (schema-functions s))

	(decls (schema-decls s))
	(preds (schema-preds s))
	(signature (schema-signature s))

	(defns (schema-defns s))
	(input (schema-input s))
	(output  (schema-output  s))
	(exists  (schema-exists  s))
	(exec (schema-exec s))
	first rest)
    (cond ((null preds) (setq first 't))
	  ((symbolp preds) (setq first preds))
	  ((eq (car preds) 'and)
	   (setq first (cadr preds))
	   (setq rest (cddr preds)))
	  (t (setq first preds)))
    (format t "~%~s ~s ~s~%[Type : ~s~% Vars : ~s~
~% Input : ~s           Output : ~s          Exists : ~s~
~% Schemas : ~s~% Functions : ~s~% Defns : ~s
~% Decls : ~s ~{~%       : ~s~} 
~% Exec  : ~s~
~% Preds : ~s ~{~%       : ~s~} ]~%"
	    kind name formals
	    signature
	    vars input output exists 
	    schemas functions defns
	    (if decls (car decls) "")
	    (if (and (consp decls) (cdr decls)) (cdr decls) "")
	    exec first rest)))

(defun reset ()
  (setq *types*
	(list (cons 'Z (make-ztype :name 'Z
				  :kind 'integer
				  :basetype 'Z
				  :fun 'integerp
				  :firstchoice 0))
	      (cons 'N (make-ztype :name 'N
				  :kind 'integer
				  :basetype 'Z
				  :fun 'integerp
				  :firstchoice 0))
	      (cons 'N1 (make-ztype :name 'N1
				  :kind 'integer
				  :basetype 'Z
				  :fun 'integerp
				  :firstchoice 1))))
  (setq *globals*      nil)
  (setq *constants*    nil)
  (setq *constructors* nil)
  (setq *functions*    nil)
  (setq *schemas*      nil)
  (setq *generics*     nil)
  (setq *axioms*       nil))

(defvar *last-file* nil)

(defun reload ()
  (load "utilities.lisp")
  (load "structures.lisp")
  (load "operators.lisp")
  (load "zlisp.lisp")
  (reset))

(defun reload-binary ()
  (load "utilities.o")
  (load "structures.o")
  (load "operators.o")
  (load "zlisp.o")
  (reset))

(defun redo (&optional (f *last-file*))
  (setq *last-file* f)
  (format *terminal-io* "~%Parsing and generating executable for ~a.~%" f)
  (reset)
  (loadz f)
  (generate-executable))

(defun check-install ()
  (reset)
  (loadz "test/lib.z")
  (generate-executable)
  (test (return-fn 'm1) (checkoutcopy-fn 'bob 'i2)(return-fn 'i2)))

